copyright = "2025 Akio Taniguchi"
extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
]
html_static_path = ["_static"]
html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "github_url": "https://github.com/astropenguin/sam45",
    "logo": {"text": "SAM45"},
    "navbar_end": [
        "version-switcher",
        "theme-switcher",
        "navbar-icon-links",
    ],
    "switcher": {
        "json_url": "https://astropenguin.github.io/sam45/_static/switcher.json",
        "version_match": "1.0.0rc1",
    },
}



# dependencies
import sam45


def test_namespace() -> None:
    sam45.dtypes.ctl
    sam45.dtypes.obs
    sam45.dtypes.dat
    sam45.dtypes.end
    sam45.read.ctl
    sam45.read.obs
    sam45.read.dat
    sam45.read.end



__all__ = ["ctl", "obs", "dat", "end"]


# standard library
from mmap import mmap
from os import PathLike
from typing import Any


# dependencies
from numpy import frombuffer
from numpy.typing import NDArray
from . import dtypes


def ctl(log: PathLike[str] | str, /, *, validate: bool = True) -> NDArray[Any]:
    """Read a SAM45 log to extract the SAM45/CTL information.

    Args:
        log: Path to the SAM45 log.
        validate: Whether to validate the content of the read information.

    Returns:
        NumPy structured array containing the SAM45/CTL information.

    Raises:
        ValueError: Raised if the read information is invalid,
            i.e., the ``crec_type`` field of it is not ``b"L0"``.

    """
    start = None
    stop = dtypes.ctl.itemsize

    with open(log, "r+b") as f:
        mm = mmap(f.fileno(), 0)
        data = frombuffer(mm[start:stop], dtype=dtypes.ctl)

    if not validate or (data["crec_type"] == b"L0").all():
        return data
    else:
        raise ValueError("Invalid SAM45/CTL information (maybe truncated).")


def obs(log: PathLike[str] | str, /, *, validate: bool = True) -> NDArray[Any]:
    """Read a SAM45 log to extract the SAM45/OBS information.

    Args:
        log: Path to the SAM45 log.
        validate: Whether to validate the content of the read information.

    Returns:
        NumPy structured array containing the SAM45/OBS information.

    Raises:
        ValueError: Raised if the read information is invalid,
            i.e., the ``crec_type`` field of it is not ``b"L1"``.

    """
    start = dtypes.ctl.itemsize
    stop = dtypes.ctl.itemsize + dtypes.obs.itemsize

    with open(log, "r+b") as f:
        mm = mmap(f.fileno(), 0)
        data = frombuffer(mm[start:stop], dtype=dtypes.obs)

    if not validate or (data["crec_type"] == b"L1").all():
        return data
    else:
        raise ValueError("Invalid SAM45/OBS information (maybe truncated).")


def dat(log: PathLike[str] | str, /, *, validate: bool = True) -> NDArray[Any]:
    """Read a SAM45 log to extract the SAM45/DAT information.

    Args:
        log: Path to the SAM45 log.
        validate: Whether to validate the content of the read information.

    Returns:
        NumPy structured array containing the SAM45/DAT information.

    Raises:
        ValueError: Raised if the read information is invalid,
            i.e., the ``crec_type`` field of it is not ``b"L2"``.

    """
    start = dtypes.ctl.itemsize + dtypes.obs.itemsize
    stop = -dtypes.end.itemsize

    with open(log, "r+b") as f:
        mm = mmap(f.fileno(), 0)
        data = frombuffer(mm[start:stop], dtype=dtypes.dat)

    if not validate or (data["crec_type"] == b"L2").all():
        return data
    else:
        raise ValueError("Invalid SAM45/DAT information (maybe truncated).")


def end(log: PathLike[str] | str, /, *, validate: bool = True) -> NDArray[Any]:
    """Read a SAM45 log to extract the SAM45/END information.

    Args:
        log: Path to the SAM45 log.
        validate: Whether to validate the content of the read information.

    Returns:
        NumPy structured array containing the SAM45/END information.

    Raises:
        ValueError: Raised if the read information is invalid,
            i.e., the ``crec_type`` field of it is not ``b"ED"``.

    """
    start = -dtypes.end.itemsize
    stop = None

    with open(log, "r+b") as f:
        mm = mmap(f.fileno(), 0)
        data = frombuffer(mm[start:stop], dtype=dtypes.end)

    if not validate or (data["crec_type"] == b"ED").all():
        return data
    else:
        raise ValueError("Invalid SAM45/END information (maybe truncated).")



__all__ = ["ctl", "obs", "dat", "end"]


# standard library
from importlib.resources import files
from tomli import loads
from typing import Any


# dependencies
import numpy as np


def _get_dtype(name: str, /) -> np.dtype[Any]:
    with (files("sam45") / "dtypes.toml").open() as f:
        text = f.read()

    consts = loads(text)["consts"]
    dtype = loads(text.format(**consts))[name]
    return np.dtype([(item["name"], item["dtype"]) for item in dtype])


ctl = _get_dtype("ctl")
"""Structured data type of the SAM45/CTL information."""

obs = _get_dtype("obs")
"""Structured data type of the SAM45/OBS information."""

dat = _get_dtype("dat")
"""Structured data type of the SAM45/DAT information."""

end = _get_dtype("end")
"""Structured data type of the SAM45/END information."""
